// Tests of installation and updates from a v2 Rust distribution
// server (mocked on the file system)

extern crate flate2;
extern crate itertools;
extern crate rustup_dist;
extern crate rustup_mock;
extern crate rustup_utils;
extern crate tar;
extern crate tempdir;
extern crate toml;
extern crate url;
extern crate walkdir;

use rustup_mock::dist::*;
use rustup_mock::{MockComponentBuilder, MockFile, MockInstallerBuilder};
use rustup_dist::prefix::InstallPrefix;
use rustup_dist::ErrorKind;
use rustup_dist::errors::Result;
use rustup_dist::dist::{TargetTriple, ToolchainDesc, DEFAULT_DIST_SERVER};
use rustup_dist::download::DownloadCfg;
use rustup_dist::Notification;
use rustup_utils::utils;
use rustup_utils::raw as utils_raw;
use rustup_dist::temp;
use rustup_dist::manifestation::{Changes, Manifestation, UpdateStatus};
use rustup_dist::manifest::{Component, Manifest};
use url::Url;
use std::cell::Cell;
use std::collections::HashMap;
use std::fs;
use std::path::Path;
use std::sync::Arc;

use tempdir::TempDir;

// Creates a mock dist server populated with some test data
pub fn create_mock_dist_server(
    path: &Path,
    edit: Option<&Fn(&str, &mut [MockPackage])>,
) -> MockDistServer {
    MockDistServer {
        path: path.to_owned(),
        channels: vec![
            create_mock_channel("nightly", "2016-02-01", edit),
            create_mock_channel("nightly", "2016-02-02", edit),
        ],
    }
}

pub fn create_mock_channel(
    channel: &str,
    date: &str,
    edit: Option<&Fn(&str, &mut [MockPackage])>,
) -> MockChannel {
    // Put the date in the files so they can be differentiated
    let contents = Arc::new(date.as_bytes().to_vec());

    let mut packages = Vec::with_capacity(5);

    packages.push(MockPackage {
        name: "rust",
        version: "1.0.0",
        targets: vec![
            MockTargetedPackage {
                target: "x86_64-apple-darwin".to_string(),
                available: true,
                components: vec![
                    MockComponent {
                        name: "rustc".to_string(),
                        target: "x86_64-apple-darwin".to_string(),
                    },
                    MockComponent {
                        name: "cargo".to_string(),
                        target: "x86_64-apple-darwin".to_string(),
                    },
                    MockComponent {
                        name: "rust-std".to_string(),
                        target: "x86_64-apple-darwin".to_string(),
                    },
                ],
                extensions: vec![
                    MockComponent {
                        name: "rust-std".to_string(),
                        target: "i686-apple-darwin".to_string(),
                    },
                    MockComponent {
                        name: "rust-std".to_string(),
                        target: "i686-unknown-linux-gnu".to_string(),
                    },
                ],
                installer: MockInstallerBuilder { components: vec![] },
            },
            MockTargetedPackage {
                target: "i686-apple-darwin".to_string(),
                available: true,
                components: vec![
                    MockComponent {
                        name: "rustc".to_string(),
                        target: "i686-apple-darwin".to_string(),
                    },
                    MockComponent {
                        name: "cargo".to_string(),
                        target: "i686-apple-darwin".to_string(),
                    },
                    MockComponent {
                        name: "rust-std".to_string(),
                        target: "i686-apple-darwin".to_string(),
                    },
                ],
                extensions: vec![],
                installer: MockInstallerBuilder { components: vec![] },
            },
        ],
    });

    for bin in &["bin/rustc", "bin/cargo"] {
        let pkg = &bin[4..];
        packages.push(MockPackage {
            name: pkg,
            version: "1.0.0",
            targets: vec![
                MockTargetedPackage {
                    target: "x86_64-apple-darwin".to_string(),
                    available: true,
                    components: vec![],
                    extensions: vec![],
                    installer: MockInstallerBuilder {
                        components: vec![
                            MockComponentBuilder {
                                name: pkg.to_string(),
                                files: vec![MockFile::new_arc(*bin, contents.clone())],
                            },
                        ],
                    },
                },
                MockTargetedPackage {
                    target: "i686-apple-darwin".to_string(),
                    available: true,
                    components: vec![],
                    extensions: vec![],
                    installer: MockInstallerBuilder { components: vec![] },
                },
            ],
        });
    }

    packages.push(MockPackage {
        name: "rust-std",
        version: "1.0.0",
        targets: vec![
            MockTargetedPackage {
                target: "x86_64-apple-darwin".to_string(),
                available: true,
                components: vec![],
                extensions: vec![],
                installer: MockInstallerBuilder {
                    components: vec![
                        MockComponentBuilder {
                            name: "rust-std-x86_64-apple-darwin".to_string(),
                            files: vec![MockFile::new_arc("lib/libstd.rlib", contents.clone())],
                        },
                    ],
                },
            },
            MockTargetedPackage {
                target: "i686-apple-darwin".to_string(),
                available: true,
                components: vec![],
                extensions: vec![],
                installer: MockInstallerBuilder {
                    components: vec![
                        MockComponentBuilder {
                            name: "rust-std-i686-apple-darwin".to_string(),
                            files: vec![
                                MockFile::new_arc(
                                    "lib/i686-apple-darwin/libstd.rlib",
                                    contents.clone(),
                                ),
                            ],
                        },
                    ],
                },
            },
            MockTargetedPackage {
                target: "i686-unknown-linux-gnu".to_string(),
                available: true,
                components: vec![],
                extensions: vec![],
                installer: MockInstallerBuilder {
                    components: vec![
                        MockComponentBuilder {
                            name: "rust-std-i686-unknown-linux-gnu".to_string(),
                            files: vec![
                                MockFile::new_arc(
                                    "lib/i686-unknown-linux-gnu/libstd.rlib",
                                    contents.clone(),
                                ),
                            ],
                        },
                    ],
                },
            },
        ],
    });

    // An extra package that can be used as a component of the other packages
    // for various tests
    packages.push(bonus_component("bonus", contents.clone()));

    if let Some(edit) = edit {
        edit(date, &mut packages);
    }

    MockChannel {
        name: channel.to_string(),
        date: date.to_string(),
        packages,
        renames: HashMap::new(),
    }
}

fn bonus_component(name: &'static str, contents: Arc<Vec<u8>>) -> MockPackage {
    MockPackage {
        name: name,
        version: "1.0.0",
        targets: vec![
            MockTargetedPackage {
                target: "x86_64-apple-darwin".to_string(),
                available: true,
                components: vec![],
                extensions: vec![],
                installer: MockInstallerBuilder {
                    components: vec![
                        MockComponentBuilder {
                            name: format!("{}-x86_64-apple-darwin", name),
                            files: vec![MockFile::new_arc("bin/bonus", contents)],
                        },
                    ],
                },
            },
        ],
    }
}

#[test]
fn mock_dist_server_smoke_test() {
    let tempdir = TempDir::new("rustup").unwrap();
    let path = tempdir.path();

    create_mock_dist_server(&path, None).write(&[ManifestVersion::V2], false);

    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rustc-nightly-x86_64-apple-darwin.tar.gz"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rustc-nightly-i686-apple-darwin.tar.gz"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rust-std-nightly-x86_64-apple-darwin.tar.gz"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rust-std-nightly-i686-apple-darwin.tar.gz"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rustc-nightly-x86_64-apple-darwin.tar.gz.sha256"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rustc-nightly-i686-apple-darwin.tar.gz.sha256"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rust-std-nightly-x86_64-apple-darwin.tar.gz.sha256"
    )));
    assert!(utils::path_exists(path.join(
        "dist/2016-02-01/rust-std-nightly-i686-apple-darwin.tar.gz.sha256"
    )));
    assert!(utils::path_exists(path.join(
        "dist/channel-rust-nightly.toml"
    )));
    assert!(utils::path_exists(path.join(
        "dist/channel-rust-nightly.toml.sha256"
    )));
}

// Test that a standard rename works - the component is installed with the old name, then renamed
// the next day to the new name.
#[test]
fn rename_component() {
    let dist_tempdir = TempDir::new("rustup").unwrap();
    let ref url = Url::parse(&format!("file://{}", dist_tempdir.path().to_string_lossy())).unwrap();

    let edit_1 = &|_: &str, pkgs: &mut [MockPackage]| {
        let tpkg = pkgs[0]
            .targets
            .iter_mut()
            .find(|p| p.target == "x86_64-apple-darwin")
            .unwrap();
        tpkg.components.push(MockComponent {
            name: "bonus".to_string(),
            target: "x86_64-apple-darwin".to_string(),
        });
    };
    let edit_2 = &|_: &str, pkgs: &mut [MockPackage]| {
        let tpkg = pkgs[0]
            .targets
            .iter_mut()
            .find(|p| p.target == "x86_64-apple-darwin")
            .unwrap();
        tpkg.components.push(MockComponent {
            name: "bobo".to_string(),
            target: "x86_64-apple-darwin".to_string(),
        });
    };

    let date_2 = "2016-02-02";
    let mut channel_2 = create_mock_channel("nightly", date_2, Some(edit_2));
    channel_2.packages[4] = bonus_component("bobo", Arc::new(date_2.as_bytes().to_vec()));
    channel_2
        .renames
        .insert("bonus".to_owned(), "bobo".to_owned());
    let mock_dist_server = MockDistServer {
        path: dist_tempdir.path().to_owned(),
        channels: vec![
            create_mock_channel("nightly", "2016-02-01", Some(edit_1)),
            channel_2,
        ],
    };

    setup_from_dist_server(
        mock_dist_server,
        url,
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            change_channel_date(url, "nightly", "2016-02-01");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
            assert!(!utils::path_exists(&prefix.path().join("bin/bobo")));
            change_channel_date(url, "nightly", "2016-02-02");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
            assert!(!utils::path_exists(&prefix.path().join("bin/bobo")));
        },
    );
}

// Test that a rename is ignored if the component with the old name was never installed.
#[test]
fn rename_component_new() {
    let dist_tempdir = TempDir::new("rustup").unwrap();
    let ref url = Url::parse(&format!("file://{}", dist_tempdir.path().to_string_lossy())).unwrap();

    let edit_2 = &|_: &str, pkgs: &mut [MockPackage]| {
        let tpkg = pkgs[0]
            .targets
            .iter_mut()
            .find(|p| p.target == "x86_64-apple-darwin")
            .unwrap();
        tpkg.components.push(MockComponent {
            name: "bobo".to_string(),
            target: "x86_64-apple-darwin".to_string(),
        });
    };

    let date_2 = "2016-02-02";
    let mut channel_2 = create_mock_channel("nightly", date_2, Some(edit_2));
    channel_2.packages[4] = bonus_component("bobo", Arc::new(date_2.as_bytes().to_vec()));
    channel_2
        .renames
        .insert("bonus".to_owned(), "bobo".to_owned());
    let mock_dist_server = MockDistServer {
        path: dist_tempdir.path().to_owned(),
        channels: vec![
            create_mock_channel("nightly", "2016-02-01", None),
            channel_2,
        ],
    };

    setup_from_dist_server(
        mock_dist_server,
        url,
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            change_channel_date(url, "nightly", "2016-02-01");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(!utils::path_exists(&prefix.path().join("bin/bonus")));
            assert!(!utils::path_exists(&prefix.path().join("bin/bobo")));
            change_channel_date(url, "nightly", "2016-02-02");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(!utils::path_exists(&prefix.path().join("bin/bobo")));
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
        },
    );
}

// Installs or updates a toolchain from a dist server.  If an initial
// install then it will be installed with the default components.  If
// an upgrade then all the existing components will be upgraded.
// FIXME: Unify this with dist::update_from_dist
fn update_from_dist(
    dist_server: &Url,
    toolchain: &ToolchainDesc,
    prefix: &InstallPrefix,
    add: &[Component],
    remove: &[Component],
    download_cfg: &DownloadCfg,
    temp_cfg: &temp::Cfg,
) -> Result<UpdateStatus> {
    update_from_dist_(
        dist_server,
        toolchain,
        prefix,
        add,
        remove,
        download_cfg,
        temp_cfg,
        false,
    )
}

fn update_from_dist_(
    dist_server: &Url,
    toolchain: &ToolchainDesc,
    prefix: &InstallPrefix,
    add: &[Component],
    remove: &[Component],
    download_cfg: &DownloadCfg,
    temp_cfg: &temp::Cfg,
    force_update: bool,
) -> Result<UpdateStatus> {
    // Download the dist manifest and place it into the installation prefix
    let ref manifest_url = make_manifest_url(dist_server, toolchain)?;
    let manifest_file = temp_cfg.new_file()?;
    utils::download_file(manifest_url, &manifest_file, None, &|_| {})?;
    let manifest_str = utils::read_file("manifest", &manifest_file)?;
    let manifest = Manifest::parse(&manifest_str)?;

    // Read the manifest to update the components
    let trip = toolchain.target.clone();
    let manifestation = Manifestation::open(prefix.clone(), trip)?;

    let changes = Changes {
        add_extensions: add.to_owned(),
        remove_extensions: remove.to_owned(),
    };

    manifestation.update(
        &manifest,
        changes,
        force_update,
        download_cfg,
        download_cfg.notify_handler.clone(),
    )
}

fn make_manifest_url(dist_server: &Url, toolchain: &ToolchainDesc) -> Result<Url> {
    let url = format!(
        "{}/dist/channel-rust-{}.toml",
        dist_server, toolchain.channel
    );

    Ok(Url::parse(&url).unwrap())
}

fn uninstall(
    toolchain: &ToolchainDesc,
    prefix: &InstallPrefix,
    temp_cfg: &temp::Cfg,
    notify_handler: &Fn(Notification),
) -> Result<()> {
    let trip = toolchain.target.clone();
    let manifestation = Manifestation::open(prefix.clone(), trip)?;
    let manifest = manifestation.load_manifest()?.unwrap();

    manifestation.uninstall(&manifest, temp_cfg, notify_handler.clone())?;

    Ok(())
}

fn setup(
    edit: Option<&Fn(&str, &mut [MockPackage])>,
    enable_xz: bool,
    f: &Fn(&Url, &ToolchainDesc, &InstallPrefix, &DownloadCfg, &temp::Cfg),
) {
    let dist_tempdir = TempDir::new("rustup").unwrap();
    let mock_dist_server = create_mock_dist_server(dist_tempdir.path(), edit);
    let ref url = Url::parse(&format!("file://{}", dist_tempdir.path().to_string_lossy())).unwrap();
    setup_from_dist_server(mock_dist_server, url, enable_xz, f);
}

fn setup_from_dist_server(
    server: MockDistServer,
    url: &Url,
    enable_xz: bool,
    f: &Fn(&Url, &ToolchainDesc, &InstallPrefix, &DownloadCfg, &temp::Cfg),
) {
    server.write(&[ManifestVersion::V2], enable_xz);

    let prefix_tempdir = TempDir::new("rustup").unwrap();

    let work_tempdir = TempDir::new("rustup").unwrap();
    let ref temp_cfg = temp::Cfg::new(
        work_tempdir.path().to_owned(),
        DEFAULT_DIST_SERVER,
        Box::new(|_| ()),
    );

    let ref toolchain = ToolchainDesc::from_str("nightly-x86_64-apple-darwin").unwrap();
    let ref prefix = InstallPrefix::from(prefix_tempdir.path().to_owned());
    let ref download_cfg = DownloadCfg {
        dist_root: "phony",
        temp_cfg: temp_cfg,
        download_dir: &prefix.path().to_owned().join("downloads"),
        notify_handler: &|_| {},
    };

    f(url, toolchain, prefix, download_cfg, temp_cfg);
}

#[test]
fn initial_install() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix.path().join("bin/rustc")));
        assert!(utils::path_exists(&prefix.path().join("lib/libstd.rlib")));
    });
}

#[test]
fn initial_install_xz() {
    setup(None, true, &|url,
                        toolchain,
                        prefix,
                        download_cfg,
                        temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix.path().join("bin/rustc")));
        assert!(utils::path_exists(&prefix.path().join("lib/libstd.rlib")));
    });
}

#[test]
fn test_uninstall() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
        uninstall(toolchain, prefix, temp_cfg, &|_| ()).unwrap();

        assert!(!utils::path_exists(&prefix.path().join("bin/rustc")));
        assert!(!utils::path_exists(&prefix.path().join("lib/libstd.rlib")));
    });
}

#[test]
fn uninstall_removes_config_file() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
        assert!(utils::path_exists(&prefix.manifest_file("multirust-config.toml")));
        uninstall(toolchain, prefix, temp_cfg, &|_| ()).unwrap();
        assert!(!utils::path_exists(&prefix.manifest_file("multirust-config.toml")));
    });
}

#[test]
fn upgrade() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        change_channel_date(url, "nightly", "2016-02-01");
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
        assert_eq!(
            "2016-02-01",
            utils_raw::read_file(&prefix.path().join("bin/rustc")).unwrap()
        );
        change_channel_date(url, "nightly", "2016-02-02");
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
        assert_eq!(
            "2016-02-02",
            utils_raw::read_file(&prefix.path().join("bin/rustc")).unwrap()
        );
    });
}

#[test]
fn unavailable_component() {
    // On day 2 the bonus component is no longer available
    let edit = &|date: &str, pkgs: &mut [MockPackage]| {
        // Require the bonus component every dat
        {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.components.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }

        // Mark the bonus package as unavailable in 2016-02-02
        if date == "2016-02-02" {
            let bonus_pkg = pkgs.iter_mut().find(|p| p.name == "bonus").unwrap();

            for target in &mut bonus_pkg.targets {
                target.available = false;
            }
        }
    };

    setup(
        Some(edit),
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            change_channel_date(url, "nightly", "2016-02-01");
            // Update with bonus.
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
            change_channel_date(url, "nightly", "2016-02-02");

            // Update without bonus, should fail.
            let err = update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg)
                .unwrap_err();
            match *err.kind() {
                ErrorKind::RequestedComponentsUnavailable(..) => {}
                _ => panic!(),
            }
        },
    );
}

#[test]
fn removed_component() {
    // On day 1 install the 'bonus' component, on day 2 its no longer a component
    let edit = &|date: &str, pkgs: &mut [MockPackage]| {
        if date == "2016-02-01" {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.components.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }
    };

    setup(
        Some(edit),
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            let received_notification = Arc::new(Cell::new(false));

            let download_cfg = DownloadCfg {
                dist_root: download_cfg.dist_root,
                temp_cfg: download_cfg.temp_cfg,
                download_dir: download_cfg.download_dir,
                notify_handler: &|n| {
                    if let Notification::ComponentUnavailable("bonus", Some(_)) = n {
                        received_notification.set(true);
                    }
                },
            };

            change_channel_date(url, "nightly", "2016-02-01");
            // Update with bonus.
            update_from_dist(url, toolchain, prefix, &[], &[], &download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
            change_channel_date(url, "nightly", "2016-02-02");

            // Update without bonus, should emit a notify and remove the bonus component
            update_from_dist(url, toolchain, prefix, &[], &[], &download_cfg, temp_cfg).unwrap();
            assert!(!utils::path_exists(&prefix.path().join("bin/bonus")));

            assert!(received_notification.get());
        },
    );
}

#[test]
fn update_preserves_extensions() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        change_channel_date(url, "nightly", "2016-02-01");
        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));

        change_channel_date(url, "nightly", "2016-02-02");
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
fn update_preserves_extensions_that_became_components() {
    let edit = &|date: &str, pkgs: &mut [MockPackage]| {
        if date == "2016-02-01" {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.extensions.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }
        if date == "2016-02-02" {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.components.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }
    };
    setup(
        Some(edit),
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            let ref adds = vec![
                Component::new(
                    "bonus".to_string(),
                    Some(TargetTriple::from_str("x86_64-apple-darwin")),
                ),
            ];

            change_channel_date(url, "nightly", "2016-02-01");
            update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));

            change_channel_date(url, "nightly", "2016-02-02");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
        },
    );
}

#[test]
fn update_preserves_components_that_became_extensions() {
    let edit = &|date: &str, pkgs: &mut [MockPackage]| {
        if date == "2016-02-01" {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.components.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }
        if date == "2016-02-02" {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.extensions.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }
    };
    setup(
        Some(edit),
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            change_channel_date(url, "nightly", "2016-02-01");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
            change_channel_date(url, "nightly", "2016-02-02");
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));
        },
    );
}

#[test]
fn update_makes_no_changes_for_identical_manifest() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let status =
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
        assert_eq!(status, UpdateStatus::Changed);
        let status =
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();
        assert_eq!(status, UpdateStatus::Unchanged);
    });
}

#[test]
fn add_extensions_for_initial_install() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
fn add_extensions_for_same_manifest() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
fn add_extensions_for_upgrade() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        change_channel_date(url, "nightly", "2016-02-01");

        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        change_channel_date(url, "nightly", "2016-02-02");

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
#[should_panic]
fn add_extension_not_in_manifest() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rust-bogus".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();
    });
}

#[test]
#[should_panic]
fn add_extension_that_is_required_component() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rustc".to_string(),
                Some(TargetTriple::from_str("x86_64-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();
    });
}

#[test]
#[ignore]
fn add_extensions_for_same_manifest_does_not_reinstall_other_components() {}

#[test]
#[ignore]
fn add_extensions_for_same_manifest_when_extension_already_installed() {}

#[test]
fn add_extensions_does_not_remove_other_components() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix.path().join("bin/rustc")));
    });
}

// Asking to remove extensions on initial install is nonsese.
#[test]
#[should_panic]
fn remove_extensions_for_initial_install() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref removes = vec![
            Component::new(
                "rustc".to_string(),
                Some(TargetTriple::from_str("x86_64-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();
    });
}

#[test]
fn remove_extensions_for_same_manifest() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();

        assert!(!utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
fn remove_extensions_for_upgrade() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        change_channel_date(url, "nightly", "2016-02-01");

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        change_channel_date(url, "nightly", "2016-02-02");

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();

        assert!(!utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
#[should_panic]
fn remove_extension_not_in_manifest() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        change_channel_date(url, "nightly", "2016-02-01");

        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        change_channel_date(url, "nightly", "2016-02-02");

        let ref removes = vec![
            Component::new(
                "rust-bogus".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();
    });
}

// Extensions that don't exist in the manifest may still exist on disk
// from a previous manifest. The can't be requested to be removed though;
// only things in the manifest can.
#[test]
#[should_panic]
fn remove_extension_not_in_manifest_but_is_already_installed() {
    let edit = &|date: &str, pkgs: &mut [MockPackage]| {
        if date == "2016-02-01" {
            let tpkg = pkgs[0]
                .targets
                .iter_mut()
                .find(|p| p.target == "x86_64-apple-darwin")
                .unwrap();
            tpkg.extensions.push(MockComponent {
                name: "bonus".to_string(),
                target: "x86_64-apple-darwin".to_string(),
            });
        }
    };
    setup(
        Some(edit),
        false,
        &|url, toolchain, prefix, download_cfg, temp_cfg| {
            change_channel_date(url, "nightly", "2016-02-01");

            let ref adds = vec![
                Component::new(
                    "bonus".to_string(),
                    Some(TargetTriple::from_str("x86_64-apple-darwin")),
                ),
            ];
            update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();
            assert!(utils::path_exists(&prefix.path().join("bin/bonus")));

            change_channel_date(url, "nightly", "2016-02-02");

            let ref removes = vec![
                Component::new(
                    "bonus".to_string(),
                    Some(TargetTriple::from_str("x86_64-apple-darwin")),
                ),
            ];
            update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();
        },
    );
}

#[test]
#[should_panic]
fn remove_extension_that_is_required_component() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        let ref removes = vec![
            Component::new(
                "rustc".to_string(),
                Some(TargetTriple::from_str("x86_64-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();
    });
}

#[test]
#[should_panic]
fn remove_extension_not_installed() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();
    });
}

#[test]
#[ignore]
fn remove_extensions_for_same_manifest_does_not_reinstall_other_components() {}

#[test]
fn remove_extensions_does_not_remove_other_components() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, &[], removes, download_cfg, temp_cfg).unwrap();

        assert!(utils::path_exists(&prefix.path().join("bin/rustc")));
    });
}

#[test]
fn add_and_remove_for_upgrade() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        change_channel_date(url, "nightly", "2016-02-01");

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        change_channel_date(url, "nightly", "2016-02-02");

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(
            url,
            toolchain,
            prefix,
            adds,
            removes,
            download_cfg,
            temp_cfg,
        ).unwrap();

        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(!utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
fn add_and_remove() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(url, toolchain, prefix, adds, &[], download_cfg, temp_cfg).unwrap();

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-unknown-linux-gnu")),
            ),
        ];

        update_from_dist(
            url,
            toolchain,
            prefix,
            adds,
            removes,
            download_cfg,
            temp_cfg,
        ).unwrap();

        assert!(utils::path_exists(&prefix
            .path()
            .join("lib/i686-apple-darwin/libstd.rlib")));
        assert!(!utils::path_exists(&prefix
            .path()
            .join("lib/i686-unknown-linux-gnu/libstd.rlib")));
    });
}

#[test]
#[should_panic]
fn add_and_remove_same_component() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap();

        let ref adds = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple-darwin")),
            ),
        ];

        let ref removes = vec![
            Component::new(
                "rust-std".to_string(),
                Some(TargetTriple::from_str("i686-apple_darwin")),
            ),
        ];

        update_from_dist(
            url,
            toolchain,
            prefix,
            adds,
            removes,
            download_cfg,
            temp_cfg,
        ).unwrap();
    });
}

#[test]
fn bad_component_hash() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let path = url.to_file_path().unwrap();
        let path = path.join("dist/2016-02-02/rustc-nightly-x86_64-apple-darwin.tar.gz");
        utils_raw::write_file(&path, "bogus").unwrap();

        let err =
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap_err();

        match *err.kind() {
            ErrorKind::ComponentDownloadFailed(_) => (),
            _ => panic!(),
        }
    });
}

#[test]
fn unable_to_download_component() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let path = url.to_file_path().unwrap();
        let path = path.join("dist/2016-02-02/rustc-nightly-x86_64-apple-darwin.tar.gz");
        fs::remove_file(&path).unwrap();

        let err =
            update_from_dist(url, toolchain, prefix, &[], &[], download_cfg, temp_cfg).unwrap_err();

        match *err.kind() {
            ErrorKind::ComponentDownloadFailed(..) => (),
            _ => panic!(),
        }
    });
}

fn prevent_installation(prefix: &InstallPrefix) {
    utils::ensure_dir_exists("installation path", &prefix.path().join("lib"), &|_| {}).unwrap();
    let install_blocker = prefix.path().join("lib").join("rustlib");
    utils::write_file("install-blocker", &install_blocker, "fail-installation").unwrap();
}

fn allow_installation(prefix: &InstallPrefix) {
    let install_blocker = prefix.path().join("lib").join("rustlib");
    utils::remove_file("install-blocker", &install_blocker).unwrap();
}

#[test]
fn reuse_downloaded_file() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        prevent_installation(prefix);

        let reuse_notification_fired = Arc::new(Cell::new(false));

        let download_cfg = DownloadCfg {
            dist_root: download_cfg.dist_root,
            temp_cfg: download_cfg.temp_cfg,
            download_dir: download_cfg.download_dir,
            notify_handler: &|n| {
                if let Notification::FileAlreadyDownloaded = n {
                    reuse_notification_fired.set(true);
                }
            },
        };

        update_from_dist(url, toolchain, prefix, &[], &[], &download_cfg, temp_cfg).unwrap_err();
        assert!(!reuse_notification_fired.get());

        allow_installation(&prefix);

        update_from_dist(url, toolchain, prefix, &[], &[], &download_cfg, temp_cfg).unwrap();

        assert!(reuse_notification_fired.get());
    })
}

#[test]
fn checks_files_hashes_before_reuse() {
    setup(None, false, &|url,
                         toolchain,
                         prefix,
                         download_cfg,
                         temp_cfg| {
        let path = url.to_file_path().unwrap();
        let target_hash = utils::read_file(
            "target hash",
            &path.join("dist/2016-02-02/rustc-nightly-x86_64-apple-darwin.tar.gz.sha256"),
        ).unwrap()[..64]
            .to_owned();
        let prev_download = download_cfg.download_dir.join(target_hash);
        utils::ensure_dir_exists("download dir", &download_cfg.download_dir, &|_| {}).unwrap();
        utils::write_file("bad previous download", &prev_download, "bad content").unwrap();
        println!("wrote previous download to {}", prev_download.display());

        let noticed_bad_checksum = Arc::new(Cell::new(false));
        let download_cfg = DownloadCfg {
            dist_root: download_cfg.dist_root,
            temp_cfg: download_cfg.temp_cfg,
            download_dir: download_cfg.download_dir,
            notify_handler: &|n| {
                if let Notification::CachedFileChecksumFailed = n {
                    noticed_bad_checksum.set(true);
                }
            },
        };

        update_from_dist(url, toolchain, prefix, &[], &[], &download_cfg, temp_cfg).unwrap();

        assert!(noticed_bad_checksum.get());
    })
}
